@@ -154,6 +154,7 @@ class PureScheduler:                                    # pylint: disable=r0902
         self.jobs.update(jobs)
         return self
 
+
     def add(self, job):
         """
         Adds a single ``Schedulable`` object;
@@ -412,55 +413,67 @@ class PureScheduler:                                    # pylint: disable=r0902
                 yield job
 
 
-    def predecessors(self, job: AbstractJob) -> set[AbstractJob]:
+    def _neighbours(self, attname: str, *starts: AbstractJob) -> set[AbstractJob]:
         """
-        returns a set of the jobs in this scheduler that `job` requires
+        returns a set of all the immediate neighbours of the starting points,
+        using either the 'required' or '_s_successors' attribute
         """
-        # just in case, intersect with self.jobs to be sure to only consider jobs in this scheduler
-        return (job.required & self.jobs)
+        neighbours = set()
+        for start in starts:
+            for next in getattr(start, attname):
+                # just in case
+                if next not in self.jobs:
+                    continue
+                if next not in neighbours:
+                    neighbours.add(next)
+        return neighbours
 
-    def successors(self, job: AbstractJob) -> Iterator[AbstractJob]:
+    def predecessors(self, *starts: AbstractJob) -> set[AbstractJob]:
         """
-        returns an iterator on the jobs in s that require on `job`
-        in no particular order
+        returns a set of all the jobs in this scheduler that any of the `starts` job requires
         """
-        return (downstream for downstream in self if job in downstream.required)
+        return self._neighbours("required", *starts)
+
+    def successors(self, *starts: AbstractJob, compute_backlinks=True) -> Iterator[AbstractJob]:
+        """
+        returns a set of all the jobs in this scheduler that require any of the `starts` job
+        """
+        if compute_backlinks:
+            self._backlinks()
+        yield from self._neighbours("_s_successors", *starts)
 
 
-    def predecessors_upstream(self, job: AbstractJob) -> set[AbstractJob]:
+    def _neighbours_closure(self, attname: str, *starts: AbstractJob) -> Iterator[AbstractJob]:
+        """
+        same as _neighbours, but the closure of that relationship
+        """
+        closure = set(self._neighbours(attname, *starts))
+        while True:
+            changes = 0
+            for start in closure.copy():
+                for next in self._neighbours(attname, start):
+                    if next not in closure:
+                        closure.add(next)
+                        changes += 1
+            if not changes:
+                break
+        return closure
+
+    def predecessors_upstream(self, *starts: AbstractJob) -> set[AbstractJob]:
         """
         returns a set of all the jobs that `job` depends on,
         either immediately or further up the execution path
         """
-        result = set(self.predecessors(job))
-        while True:
-            changes = 0
-            for u1 in result.copy():
-                for u2 in self.predecessors(u1):
-                    if u2 not in result:
-                        changes += 1
-                        result.add(u2)
-            if not changes:
-                break
-        return result
+        return self._neighbours_closure("required", *starts)
 
-
-    def successors_downstream(self, job: AbstractJob) -> set[AbstractJob]:
+    def successors_downstream(self, *starts: AbstractJob, compute_backlinks=True) -> set[AbstractJob]:
         """
         return a set of all the jobs that depend on `job`,
         either immediately or further down the execution path
         """
-        result = set(self.successors(job))
-        while True:
-            changes = 0
-            for d1 in result.copy():
-                for d2 in self.successors(d1):
-                    if d2 not in result:
-                        changes += 1
-                        result.add(d2)
-            if not changes:
-                break
-        return result
+        if compute_backlinks:
+            self._backlinks()
+        return self._neighbours_closure("_s_successors", *starts)
 
 
     def bypass_and_remove(self, job: AbstractJob) -> None:
@@ -495,6 +508,16 @@ class PureScheduler:                                    # pylint: disable=r0902
         self.jobs.remove(job)
 
 
+    def keep_only(self, remains: Iterable[Schedulable]) -> None:
+        """
+        modifies the scheduler, and keep only the jobs mentioned in remains in the scheduler
+
+        any job not belonging in self are ignored
+        """
+        self.jobs &= set(remains)
+        self.sanitize()
+
+
     def keep_only_between(self, *,
                           starts:Iterable[Schedulable]=None,
                           ends:Iterable[Schedulable]=None,
@@ -502,6 +525,13 @@ class PureScheduler:                                    # pylint: disable=r0902
                           keep_ends=True) -> None:
         """
         allows to select a subset of the scheduler, which is modified in place
+
+        the algorithm works as follows:
+        we preserve the jobs that are reachable (successors) of (any of) the starts vertices
+        AND that can reach (predecessors) (any of) the ends vertices
+
+        of course if starts is empty, then only the ends criteria is used
+        (since otherwise we would retain nothing); and same if ends is empty
         """
         # optional parameters
         starts = starts if starts is not None else set()
@@ -511,18 +541,22 @@ class PureScheduler:                                    # pylint: disable=r0902
         starts = set(starts)
         ends = set(ends)
 
-        to_remove = set()
+        # if one of the milestones is empty (e.g. starts is empty)
+        # it means no additional constraint,
+        # so we take the full set of nodes in that case
+        downwards = self.successors_downstream(*starts) if starts else self.jobs
+        upwards = self.predecessors_upstream(*ends) if ends else self.jobs
 
-        for start in starts:
-            to_remove.update(self.predecessors_upstream(start))
-        if not keep_starts:
-            to_remove.update(starts)
-        for end in ends:
-            to_remove.update(self.successors_downstream(end))
-        if not keep_ends:
-            to_remove.update(ends)
-        for j in to_remove:
-            self.bypass_and_remove(j)
+        preserved = downwards & upwards
+        if keep_starts:
+            preserved.update(starts)
+        if keep_ends:
+            preserved.update(ends)
+
+        # no need to replug anything, let's do it the rough way,
+        # and sanitize to remove dangling references
+        self.jobs = preserved
+        self.sanitize()
 
 
     def _entry_csv(self):
